// This may look like C code, but it is really -*- C++ -*-
/****************************************************************************
** $Id: fancycache.h 87 2007-12-18 17:11:32Z wglas $
**
** Copyright (C) 2005 EV-i Informationstechnologie GmbH
**
** Author: Wolfgang Glas / EV-i
**
** A thread-safe application cache of database or other user objects.
** 
*****************************************************************************/

#ifndef FANCY_CACHE_H
#define FANCY_CACHE_H

#include <fancyexception.h>
#include <typeinfo>
#include <string>

namespace Fancy
{
  class CacheObject;

  class FANCYBASE_API CacheKey
  {
    size_t hash;
  public:
    CacheKey(size_t h) : hash(h) {}
    virtual ~CacheKey();

    virtual CacheObject* fetch() const;

    size_t getHash() const { return hash; }
    virtual std::string getDescription() const;

    virtual bool operator== (const CacheKey &k) const = 0;

    static size_t makeHash(const unsigned char *data, size_t data_len, size_t start=0);
  };

  class FANCYBASE_API IdCacheKey : public CacheKey
  {
  public:
    IdCacheKey(const std::string &_id);
    ~IdCacheKey();

    static size_t makeHash(const std::string &_id, size_t start=0);

    std::string getDescription() const;

    const std::string &getId() const { return id; }

    bool operator== (const CacheKey &k) const;

  private:
    std::string id;
  };

  class FANCYBASE_API CacheObject
  {
  public:
    CacheObject() {}
    virtual ~CacheObject();

    virtual CacheObject* dup() const = 0;

    virtual const CacheKey &getKey() const = 0;

    virtual void validate();
    virtual void invalidate();
  private:
  };

  class FANCYBASE_API IdCacheObject : public CacheObject
  {
  public:
    IdCacheObject(const IdCacheKey &id);
    IdCacheObject(const std::string &id);
    ~IdCacheObject();

    const std::string &getId() const { return ik.getId(); }

    const CacheKey &getKey() const;
  private:
    IdCacheKey ik;
  };

  class CachePrivate;
  class CacheEntryPrivate;

  class FANCYBASE_API Cache
  {
  public:
    class FANCYBASE_API reference
    {
    public:
      reference (CacheEntryPrivate *_d, CacheObject *_obj);
      reference (const reference &r) : d(r.d),obj(r.obj) { r.d = 0; r.obj = 0; }
      virtual ~reference ();

      reference &operator=(const reference &r);

      virtual void commit();

      CacheObject *operator->() const
      {
        if (!obj) throw Exception(FANCYBASE_LOG_CACHE,LOG_ERR,"Access to null reference.");
        return obj;
      }
      CacheObject *get() const { return obj; }
    private:

      mutable CacheEntryPrivate *d;
      mutable CacheObject *obj;
    };

    template <class T> class typed_ref : public reference
    {
    public:
      typed_ref (const reference &r) : reference(r) { t=dynamic_cast<T*>(reference::get()); }
      typed_ref (const typed_ref &r) : reference(r) { t=r.t; r.t=0; }
      ~typed_ref () {}

      typed_ref &operator=(const typed_ref &r)
      { reference::operator=(r); t=r.t; r.t=0; return *this; }

      void commit() { reference::commit(); t=0; }

      T *operator->() const
      {
        if (!t) throw Exception(FANCYBASE_LOG_CACHE,LOG_ERR,"Access to null reference.");
        return t;
      }
      T *get() const { return t; }
    private:
      mutable T* t;
    };

    class FANCYBASE_API const_reference
    {
    public:
      const_reference (CacheEntryPrivate *_d, const CacheObject *_obj);
      const_reference (const const_reference &r) : d(r.d),obj(r.obj) { r.d = 0; r.obj = 0; }
      virtual ~const_reference ();

      const_reference &operator=(const const_reference &r);

      const CacheObject *operator->() const
      {
        if (!obj) throw Exception(FANCYBASE_LOG_CACHE,LOG_ERR,"Access to null reference.");
        return obj;
      }
      const CacheObject *get() const { return obj; }

    private:
      mutable CacheEntryPrivate *d;
      mutable const CacheObject *obj;
    };

    template <class T> class const_typed_ref : public const_reference
    {
    public:
      const_typed_ref (const const_reference &r) : const_reference(r) { t=dynamic_cast<const T*>(const_reference::get()); }
      const_typed_ref (const const_typed_ref &r) : const_reference(r) { t=r.t; r.t=0; }
      ~const_typed_ref () {}

      const_typed_ref &operator=(const const_typed_ref &r)
      { const_reference::operator=(r); t=r.t; r.t=0; return *this; }

      const T *operator->() const
      {
        if (!t) throw Exception(FANCYBASE_LOG_CACHE,LOG_ERR,"Access to null reference.");
        return t;
      }
      const T *get() const { return t; }
    private:
      mutable const T* t;
    };

    Cache();
    ~Cache();

    void reserve(size_t max_elements, size_t max_threads);
    
    void insert(CacheObject *o);
    const_reference insertRead(CacheObject *o);
    reference insertWrite(CacheObject *o);

    reference acquireWrite(const CacheKey &key);
    const_reference acquireRead(const CacheKey &key);
    
    void erase(const CacheKey &key);

  private:
    CachePrivate *d;
  };

}; // namespace Fancy

#endif // ! FANCY_CACHE_H
