// This may look like C code, but it is really -*- C++ -*-
/****************************************************************************
** $Id: fancyobject.h 209 2009-03-28 20:46:54Z wglas $
**
** Copyright (C) 2009 EV-i Informationstechnologie GmbH
**
** Author: Wolfgang Glas / EV-i
**
** A thread-safe reference counted object.
** 
*****************************************************************************/

#ifndef FANCYBASE_OBJECT_H
#define FANCYBASE_OBJECT_H

#include <fancybase.h>
#include <iosfwd>
#ifdef __GNUG__
#include <bits/atomic_word.h>
#endif

namespace Fancy {

#ifdef WIN32
  typedef long ref_count_t;
#elif defined ( __GNUG__ )
  typedef _Atomic_word ref_count_t;
#else
  typedef unsigned ref_count_t;
#endif

  class FANCYBASE_API Object
  {
    ref_count_t nref;

  public:
    Object() : nref(1) {}

    /*! Increment reference count by one. */
    void ref() throw();

    /*! Decrement reference count by one.
        If reference count drops to zero, delete this and return true. */
    bool deref() throw();

    virtual std::string toString() const;

  protected:
    virtual ~Object();

  private:
    Object(const Object &a);
    Object &operator= (const Object &a);
  };

  FANCYBASE_API std::ostream& operator<<(std::ostream& os, const Object &x);
  FANCYBASE_API std::ostream& writeObjectPtr(std::ostream& os, const Object* x);
  FANCYBASE_API std::string objectPtrToString(const Object* x);
 
  // Get a std::string of an ObjPtr.
#define FANCY_OBJ_PTR_STRING(p) (::Fancy::objectPtrToString(p.get()))
  // pass an ObjPtr instance to a %s format specifier.
#define FANCY_OBJ_PTR_ARG(p) (::Fancy::objectPtrToString(p.get()).c_str())

  // A pointer to a subclass of Object. For a non-intrusive approach see
  // Fancy::Ptr<T> in fancyptr.h
  template <class T> class ObjPtr
  {
    T *ref;

  public:
    ObjPtr() : ref(0) {}
    ObjPtr(T *_t) : ref(_t) {}

    template<class U>ObjPtr(const ObjPtr<U> &a) : ref(dynamic_cast<T*>(a.get()))
    { if (ref) ref->ref(); }

    ObjPtr(const ObjPtr &a) : ref(a.ref) { if (ref) ref->ref(); }

    ObjPtr &operator=(const ObjPtr &a) { 
      if (ref) ref->deref(); ref=a.ref; if (ref) ref->ref();
      return *this;
    }

    template<class U>ObjPtr &operator=(const ObjPtr<U> &a) { 
      if (ref) ref->deref();
      ref=dynamic_cast<T*>(a.get());
      if (ref) ref->ref();
      return *this;
    }

    ObjPtr &operator=(T *_t) {
      if (ref) ref->deref();
      ref=_t;
      return *this;
    }

    void setNull() { if (ref) ref->deref(); ref=0; }

    ~ObjPtr() { if (ref) ref->deref(); }

    bool isNull() const { return ref == 0; }
    T *operator->() const { return ref; }
    T &operator*() const { return *ref; }
    
    /*!
       Indirection of pointer.
       This method should normally not be used, because
       the lifetime of returned pointer is tied to the lifetime of this
       Ptr object.
     */
    T *get() const { return ref; }

    bool operator!() const {
      return ref == 0;
    }

    template<class U> bool operator== (const ObjPtr<U> &a) {
      return a.get() == this->get();
    }

    template<class U> bool operator!= (const ObjPtr<U> &a) {
      return a.get() != this->get();
    }

    template<class U> bool operator< (const ObjPtr<U> &a) {
      return a.get() < this->get();
    }
  };
  
  template<class T> 
  inline std::ostream& operator<<(std::ostream& os, const ObjPtr<T> &p) {
    
    return writeObjectPtr(os,p.get());
  }

}

#endif
