// This may look like C code, but it is really -*- C++ -*-
/****************************************************************************
** $Id: fancythread.h 201 2009-03-18 12:53:39Z wglas $
**
** Copyright (C) 2003 EV-i Informationstechnologie GmbH
**
** Author: Wolfgang Glas / EV-i
**
** Wrapper calsses for pthread functions.
** 
*****************************************************************************/

#ifndef FANCYBASE_THREAD_H
#define FANCYBASE_THREAD_H

#include <fancybase.h>

#ifdef WIN32
typedef void *pthread_mutex_t;
namespace Fancy { class WaitConditionPrivate; }
typedef Fancy::WaitConditionPrivate *pthread_cond_t;
typedef void *pthread_t;
typedef unsigned int pthread_key_t;
#else
#include <pthread.h>
#endif

namespace Fancy {

  class FANCYBASE_API Mutex
  {
  public:
    Mutex() throw();
    ~Mutex() throw();
    
    void lock() throw();
    void unlock() throw();

    pthread_mutex_t *handle() throw();
  private:
    Mutex(const Mutex &);
    Mutex &operator= (const Mutex &);

    pthread_mutex_t mutex; 
  };

  class FANCYBASE_API FileMutex
  {
  public:
    FileMutex();
    ~FileMutex();
    
    void open(const char *fn);

    void lock();
    void unlock();

  private:
    FileMutex(const FileMutex &);
    FileMutex &operator= (const FileMutex &);

    int fd;
  };

  template<class MutexT>
  class BasicMutexLocker
  {
    MutexT *mutex;
  public:
    BasicMutexLocker(MutexT *m) : mutex(m) { if (mutex) mutex->lock(); }
    ~BasicMutexLocker() { if (mutex) mutex->unlock(); }
    void unlock() { if (mutex) { mutex->unlock(); mutex=0; } }
    void lock(MutexT *m)
    { if (mutex) mutex->unlock(); mutex=m; if (mutex) mutex->lock(); }
    MutexT *stealLock()
    { MutexT *ret = mutex; mutex = 0; return ret; }
  private:
    BasicMutexLocker(const BasicMutexLocker &);
    BasicMutexLocker &operator= (const BasicMutexLocker &);
  };

  typedef BasicMutexLocker<Mutex> MutexLocker;
  typedef BasicMutexLocker<FileMutex> FileMutexLocker;

  class FANCYBASE_API WaitCondition
  {
  public:
    WaitCondition();
    ~WaitCondition();
    
    void broadcast();
    void signal();

    void wait (Mutex *mutex);
    bool timedWait(Mutex *mutex, unsigned msec);
  private:
    pthread_cond_t cond;
  };

  class FANCYBASE_API RWMutex
  {
  public:
    RWMutex();
    ~RWMutex();

    void rLock();
    void rUnlock();

    bool wLockTimed(int msec);
    void wLock();
    void wUnlock();

  private:
    int n_read;
    Mutex n_read_mutex;
    WaitCondition n_read_wc;
    Mutex write_mutex;
  };

  class FANCYBASE_API RMutexLocker
  {
    RWMutex *mutex;
  public:
    RMutexLocker(RWMutex *m) : mutex(m) { if (mutex) mutex->rLock(); }
    ~RMutexLocker() { if (mutex) mutex->rUnlock(); }
    void unlock() { if (mutex) { mutex->rUnlock(); mutex=0; } }
    void lock(RWMutex *m)
    { if (mutex) mutex->rUnlock(); mutex=m; if (mutex) mutex->rLock(); }
  private:
    RMutexLocker(const RMutexLocker &);
    RMutexLocker &operator= (const RMutexLocker &);
  };

  class FANCYBASE_API WMutexLocker
  {
    RWMutex *mutex;
  public:
    WMutexLocker() : mutex(0) {}
    WMutexLocker(RWMutex *m) : mutex(m) { if (mutex) mutex->wLock(); }
    ~WMutexLocker() { if (mutex) mutex->wUnlock(); }
    void unlock() { if (mutex) { mutex->wUnlock(); mutex=0; } }
    void lock(RWMutex *m)
    { if (mutex) mutex->wUnlock(); mutex=m; if (mutex) mutex->wLock(); }
    bool isLocked() const { return mutex!=0; }
    bool lockTimed(RWMutex *m, int msec)
    { if (mutex) mutex->wUnlock(); mutex=0; if (m && m->wLockTimed(msec)) mutex = m; return isLocked(); }
  private:
    WMutexLocker(const WMutexLocker &);
    WMutexLocker &operator= (const WMutexLocker &);
  };

  class FANCYBASE_API Thread
  {
    friend class ThreadStarter;

  public:
#ifdef WIN32
    typedef unsigned retval_t;
#else
    typedef void *retval_t;
#endif

    class FANCYBASE_API Callback
    {
    public:
      Callback();

      virtual ~Callback();

      virtual void start(void) = 0;
      virtual void leave(void) = 0;
    };

    class FANCYBASE_API Specific
    {
    public:
      Specific();
      virtual ~Specific();

      static Specific *getSpecific(pthread_key_t key);
      static void setSpecific(pthread_key_t key, Specific *v);
    };

    class FANCYBASE_API SpecificFactory
    {
    public:
      SpecificFactory();

      virtual ~SpecificFactory();

      virtual Specific *newSpecific() = 0;
    };

    Thread();
    virtual ~Thread();

    void kill(int sig);
    
    retval_t wait();

    void start();
    bool isSelf() const;

    bool isRunning() const { return running; }

    pthread_t handle() const;
    static pthread_t currentHandle();

#ifdef WIN32
    static void *getSignalWinEventHandle();
    static void *getIOWinEventHandle();
    static int getSignalCatched();
#endif
    static void msleep (unsigned msec);
    static void registerCallback (Callback *cb);

    static pthread_key_t generateSpecificKey(SpecificFactory *factory);

  protected:
    virtual void run() = 0;

  private:
    bool running;
    pthread_t thread;
  };

#define FANCYBASE_THREAD_SUCCESS          ((::Fancy::Thread::retval_t)0)
#define FANCYBASE_THREAD_EXCEPTION        ((::Fancy::Thread::retval_t)-1)
#define FANCYBASE_THREAD_STD_EXCEPTION    ((::Fancy::Thread::retval_t)-2)
#define FANCYBASE_THREAD_UNKNOW_EXCEPTION ((::Fancy::Thread::retval_t)-3)
#define FANCYBASE_THREAD_ERRNO(a)         ((int)(a))

}

#endif // ! FANCYBASE_THREAD_H
